// crystalline-cortex.c (C) 2006,2013 Claude Heiland-Allen <claude@mathr.co.uk>
// License: GPLv2+ https://www.gnu.org/licenses/gpl-2.0.html

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define W 1024
#define H 576
#define X 128
#define Y 288

struct hist { int level, count; };
static struct hist histogram[256];

static int cmp_level(const void *a, const void *b) {
  const struct hist *x = a;
  const struct hist *y = b;
  return x->level - y->level;
}

static int cmp_count(const void *a, const void *b) {
  const struct hist *x = a;
  const struct hist *y = b;
  return y->count - x->count;
}

static unsigned char buffer[2][H][W];
static unsigned char in[8][H][W][3];
static const int kernel[3][3] = { { 1, 1, 1 }, { 1, 0, 1 } , { 1, 1, 1 } };
static unsigned char out[3][H][W];

/* {{{ BEGIN colorspace.c {{{ */
/*
 * colorspace.c:  Routines to perform colorspace conversions.
 *
 *
 *  Copyright (C) 2001 Matthew J. Marjanovic <maddog@mir.com>
 *
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */

/* precomputed tables */
#define FP_BITS 18
static int Y_R[256];
static int Y_G[256];
static int Y_B[256];
static int Cb_R[256];
static int Cb_G[256];
static int Cb_B[256];
static int Cr_R[256];
static int Cr_G[256];
static int Cr_B[256];
static int conv_RY_inited = 0;

static inline int myround(double n) {
  if (n >= 0) return (int)(n + 0.5);
  else return (int)(n - 0.5);
}

static void init_RGB_to_YCbCr_tables(void) {
  /*
   * Q_Z[i] =   (coefficient * i
   *             * (Q-excursion) / (Z-excursion) * fixed-point-factor)
   *
   * to one of each, add the following:
   *             + (fixed-point-factor / 2)         --- for rounding later
   *             + (Q-offset * fixed-point-factor)  --- to add the offset
   *             
   */
  for (int i = 0; i < 256; i++) {
    Y_R[i]  = myround(0.299 * (double)i * 219.0 / 255.0 * (double)(1<<FP_BITS));
    Y_G[i]  = myround(0.587 * (double)i * 219.0 / 255.0 * (double)(1<<FP_BITS));
    Y_B[i]  = myround((0.114 * (double)i * 219.0 / 255.0 * (double)(1<<FP_BITS))
            + (double)(1<<(FP_BITS-1)) + (16.0 * (double)(1<<FP_BITS)));
    Cb_R[i] = myround(-0.168736 * (double)i * 224.0 / 255.0 * (double)(1<<FP_BITS));
    Cb_G[i] = myround(-0.331264 * (double)i * 224.0 / 255.0 * (double)(1<<FP_BITS));
    Cb_B[i] = myround((0.500 * (double)i * 224.0 / 255.0 * (double)(1<<FP_BITS))
            + (double)(1<<(FP_BITS-1)) + (128.0 * (double)(1<<FP_BITS)));
    Cr_R[i] = myround(0.500 * (double)i * 224.0 / 255.0 * (double)(1<<FP_BITS));
    Cr_G[i] = myround(-0.418688 * (double)i * 224.0 / 255.0 * (double)(1<<FP_BITS));
    Cr_B[i] = myround((-0.081312 * (double)i * 224.0 / 255.0 * (double)(1<<FP_BITS))
            + (double)(1<<(FP_BITS-1)) + (128.0 * (double)(1<<FP_BITS)));
  }
  conv_RY_inited = 1;
}

static void rgb_to_yuv_inplace(int count, unsigned char *ptr) {
  for (int i = 0; i < count; i += 3) {
    int r = ptr[i+0];
    int g = ptr[i+1];
    int b = ptr[i+2];
    ptr[i+0] = ( Y_R[r] +  Y_G[g]+  Y_B[b]) >> FP_BITS;
    ptr[i+1] = (Cb_R[r] + Cb_G[g]+ Cb_B[b]) >> FP_BITS;
    ptr[i+2] = (Cr_R[r] + Cr_G[g]+ Cr_B[b]) >> FP_BITS;
  }
}
/* }}} END   colorspace.c }}} */

int main(int argc, char **argv) {
  (void) argc;
  (void) argv;
  int which = 0;
  int add = 0;
  int and = 0;
  memset(&buffer[0][0][0], 0, 2 * H * W);
  memset(&out[0][0][0], 0, 3 * H * W);
  for (int i = 0; i < 8; ++i) {
    int w = 0, h = 0;
    if (2 != fscanf(stdin, "P6\n%d %d 255\n", &w, &h)) return 1;
    if (w != W || h != H) return 1;
    if (1 != fread(&in[i][0][0][0], 3 * W * H, 1, stdin)) return 1;
  }
  init_RGB_to_YCbCr_tables();
  rgb_to_yuv_inplace(8 * W * H * 3, &in[0][0][0][0]);
  printf("YUV4MPEG2 W1024 H576 F25:1 Ip A1:1 C444\n");
  for (int frame = 0; frame < 90000; ++frame) {
    if (frame % 15 == 0) {
      add = ((rand() >> 16) & 255) & 63;
      and = ((rand() >> 16) & 255) | 192;
    }
    buffer[1 - which][Y][X*(2*(frame%4)+1)] ^= 255;
    #pragma omp parallel for
    for (int y = 0; y < H; ++y) {
      for (int x = 0; x < W; ++x) {
        int s = 0;
        for (int dy = -1; dy <= 1; ++dy) {
          int yy = (y + dy + H) % H;
          for (int dx = -1; dx <= 1; ++dx) {
            int xx = (x + dx + W) % W;
            s += kernel[dy+1][dx+1] * buffer[1 - which][yy][xx];
          }
        }
        buffer[which][y][x] = ((s >> 3) + add) & and;
      }
    }
    for (int i = 0; i < 256; ++i) {
      histogram[i].level = i;
      histogram[i].count = 0;
    }
    for (int y = 0; y < H; ++y) {
      for (int x = 0; x < W; ++x) {
        histogram[buffer[which][y][x]].count++;
      }
    }
    qsort(histogram, 256, sizeof(*histogram), cmp_count);
    for (int i = 0; i < 256; ++i) {
      histogram[i].count = i <= 7 ? i : 7;
    }
    qsort(histogram, 256, sizeof(*histogram), cmp_level);
    #pragma omp parallel for
    for (int y = 0; y < H; ++y) {
      for (int x = 0; x < W; ++x) {
        int i = histogram[buffer[which][y][x]].count;
        out[0][y][x] = in[i][y][x][0];
        out[2][y][x] = in[i][y][x][1];
        out[1][y][x] = in[i][y][x][2];
      }
    }
    printf("FRAME\n");
    fwrite(&out[0][0][0], H * W * 3, 1, stdout);
    which = 1 - which;
  }
  return 0;
}
